#include "storage.h"
#include "nrf_delay.h"

#define NRF_LOG_MODULE_NAME "STORAGE"
#define NRF_LOG_LEVEL 4

#include "nrf_log.h"
#include "nrf_log_ctrl.h"

NRF_QUEUE_DEF(storage_operation_t, m_operation_queue, 10,
              NRF_QUEUE_MODE_NO_OVERFLOW);

static bool is_writing = false;
static bool is_deleting = false;
static bool is_collecting = false;

static storage_operation_t m_current_operation = {0};

static ret_code_t _storage_fds_write(storage_key_t key, void *p_data,
                                     uint16_t length_words) {
  NRF_LOG_DEBUG("_storage_fds_write\n");
  fds_record_t record = {0};
  fds_record_desc_t record_desc = {0};
  fds_record_chunk_t record_chunk = {0};
  fds_find_token_t ftok = {0};

  nrf_delay_ms(10);
  NRF_LOG_HEXDUMP_DEBUG(p_data, length_words*(sizeof(uint32_t)));
  nrf_delay_ms(10);

  record_chunk.p_data = p_data;
  record_chunk.length_words = length_words;

  record.file_id = STORAGE_FILE_ID;
  record.key = key;
  record.data.p_chunks = &record_chunk;
  record.data.num_chunks = 1;

  ret_code_t err_code = fds_record_find(STORAGE_FILE_ID, key, &record_desc, &ftok);

  if (err_code == FDS_SUCCESS) {
    NRF_LOG_DEBUG("Record Exist\r\n");
    err_code = fds_record_update(&record_desc, &record);
  } else if (err_code == FDS_ERR_NOT_FOUND) {
    NRF_LOG_DEBUG("Record Does Not Exist\r\n");
    err_code = fds_record_write(&record_desc, &record);
  }

  if (err_code != FDS_SUCCESS) {
    if (err_code == FDS_ERR_NO_SPACE_IN_FLASH) {
      NRF_LOG_DEBUG("No Space in Flash\n");
    } else {
      NRF_LOG_ERROR("fds_record_write fail: %d\n", err_code);
    }
    return err_code;
  }

  return NRF_SUCCESS;
}

static ret_code_t _storage_fds_read(storage_key_t key, void *p_data,
                                    uint16_t length_words) {
  ret_code_t err_code;
  fds_flash_record_t flash_record = {0};
  fds_record_desc_t record_desc = {0};
  fds_find_token_t ftok = {0};

  err_code = fds_record_find(STORAGE_FILE_ID, key, &record_desc, &ftok);

  if (err_code != FDS_SUCCESS) {
    if (err_code != FDS_ERR_NOT_FOUND) {
      NRF_LOG_ERROR("fds_record_find fail\n", err_code);
    }
    return err_code;
  }

  err_code = fds_record_open(&record_desc, &flash_record);

  if (err_code != FDS_SUCCESS) {
    NRF_LOG_ERROR("fds_record_open fail: %d\n", err_code);
    return err_code;
  }

  memcpy(p_data, flash_record.p_data, length_words);

  err_code = fds_record_close(&record_desc);

  if (err_code != FDS_SUCCESS) {
    NRF_LOG_ERROR("fds_record_close fail: %d\n", err_code);
    return err_code;
  }

  return NRF_SUCCESS;
}

static ret_code_t _storage_fds_delete(storage_key_t key) {
  ret_code_t err_code;
  fds_record_desc_t record_desc;
  fds_find_token_t ftok = {0};

  err_code = fds_record_find(STORAGE_FILE_ID, key, &record_desc, &ftok);

  if (err_code != FDS_SUCCESS) {
    if (err_code != FDS_ERR_NOT_FOUND) {
      NRF_LOG_ERROR("fds_record_find fail: %d\n", err_code);
    }
    return err_code;
  }

  err_code = fds_record_delete(&record_desc);

  if (err_code != FDS_SUCCESS) {
    NRF_LOG_ERROR("fds_record_delete fail: %d\n", err_code);
    return err_code;
  }

  return NRF_SUCCESS;
}

static ret_code_t _storage_perform_writing_operation(void) {
  is_writing = true;
  ret_code_t err_code;
  uint16_t length_words = 0;

  switch (m_current_operation.type) {
    case STORAGE_WRITE_CALIBRATIONS_OPERATION: {
      if ((sizeof(m_current_operation.data.calibration_values) %
           sizeof(uint32_t)) == 0) {
        length_words = sizeof(m_current_operation.data.calibration_values) /
                       sizeof(uint32_t);
      } else {
        length_words = (sizeof(m_current_operation.data.calibration_values) /
                        sizeof(uint32_t)) +
                       1;
      }
      err_code = _storage_fds_write(STORAGE_CALIBRATIONS_KEY,
                                    m_current_operation.data.calibration_values,
                                    length_words);
      break;
    }
    case STORAGE_WRITE_DOSE_TIMES_OPERATION: {
      if ((sizeof(m_current_operation.data.dose_times) % sizeof(uint32_t)) ==
          0) {
        length_words =
            sizeof(m_current_operation.data.dose_times) / sizeof(uint32_t);
      } else {
        length_words =
            (sizeof(m_current_operation.data.dose_times) / sizeof(uint32_t)) +
            1;
      }
      err_code =
          _storage_fds_write(STORAGE_DOSE_TIMES_KEY,
                             m_current_operation.data.dose_times, length_words);
      break;
    }
    case STORAGE_WRITE_COULOMB_COUNTER_OPERATION: {
      if ((sizeof(m_current_operation.data.coulomb_counter_data) %
           sizeof(uint32_t)) == 0) {
        length_words = sizeof(m_current_operation.data.coulomb_counter_data) /
                       sizeof(uint32_t);
      } else {
        length_words = (sizeof(m_current_operation.data.coulomb_counter_data) /
                        sizeof(uint32_t)) +
                       1;
      }

      err_code = _storage_fds_write(
          STORAGE_CC_KEY, &(m_current_operation.data.coulomb_counter_data),
          length_words);
      break;
    }
    default: { return NRF_ERROR_INVALID_STATE; }
  }

  if (err_code != FDS_SUCCESS) {
    // If there's no space in flash, we call the garbage collector. The current
    // operation will be perform again in the fds event handler, when the
    // garbage collection proccess is finished.
    if (err_code == FDS_ERR_NO_SPACE_IN_FLASH) {
      is_collecting = true;
      is_writing = false;
      err_code = fds_gc();
      if (err_code) {
        NRF_LOG_ERROR("fds_gc fail: %d\n", err_code);
        return err_code;
      }
    } else {
      NRF_LOG_ERROR("Error Writing In Memory: %d\n", err_code);
      return err_code;
    }
  }

  return NRF_SUCCESS;
}

static ret_code_t _storage_perform_deleting_operation(void) {
  is_deleting = true;
  ret_code_t err_code;

  if (m_current_operation.type == STORAGE_DELETE_CALIBRATIONS_OPERATION) {
    err_code = _storage_fds_delete(STORAGE_CALIBRATIONS_KEY);
  } else if (m_current_operation.type == STORAGE_DELETE_DOSE_TIMES_OPERATION) {
    err_code = _storage_fds_delete(STORAGE_DOSE_TIMES_KEY);
  } else {
    return NRF_ERROR_INVALID_STATE;
  }

  if (err_code != FDS_SUCCESS) {
    if (err_code == FDS_ERR_NOT_FOUND) {
      NRF_LOG_ERROR("Record not found\n");
    } else {
      NRF_LOG_ERROR("_storage_fds_delete fail: %d\n", err_code);
    }

    return err_code;
  }

  return NRF_SUCCESS;
}

static ret_code_t _storage_perform_current_operation(void) {
  ret_code_t err_code = NRF_ERROR_INVALID_STATE;

  if (m_current_operation.type == STORAGE_WRITE_CALIBRATIONS_OPERATION ||
      m_current_operation.type == STORAGE_WRITE_DOSE_TIMES_OPERATION ||
      m_current_operation.type == STORAGE_WRITE_COULOMB_COUNTER_OPERATION) {
    err_code = _storage_perform_writing_operation();

    if (err_code) {
      NRF_LOG_ERROR("_storage_perform_writing_operation fail: %d\n", err_code);
      return err_code;
    }
  } else if (m_current_operation.type ==
                 STORAGE_DELETE_CALIBRATIONS_OPERATION ||
             m_current_operation.type == STORAGE_DELETE_DOSE_TIMES_OPERATION) {
    err_code = _storage_perform_deleting_operation();
    if (err_code) {
      NRF_LOG_ERROR("_storage_perform_deleting_operation fail: %d\n", err_code);
      return err_code;
    }
  }

  return NRF_SUCCESS;
}

static ret_code_t _storage_push_operation(storage_operation_t *p_operation) {
  if (p_operation == NULL) {
    NRF_LOG_ERROR("Invalid Null Pointer\n");
    return NRF_ERROR_NULL;
  }

  ret_code_t err_code = NRF_ERROR_INVALID_STATE;

  if (nrf_queue_is_full(&m_operation_queue)) {
    return FDS_ERR_NO_SPACE_IN_QUEUES;
  }

  if (nrf_queue_is_empty(&m_operation_queue)) {
    // If is writing, deleting or collecting we push the operation in the queue.
    if (is_writing || is_deleting || is_collecting) {
      NRF_LOG_DEBUG("Pushing Operation in Queue\n");
      NRF_LOG_DEBUG("Operation Data\n");
      NRF_LOG_HEXDUMP_DEBUG(&(p_operation->data), sizeof(p_operation->data));
      
      err_code = nrf_queue_push(&m_operation_queue, p_operation);

      if (err_code) {
        NRF_LOG_ERROR("queue_push fail: %lu\n", err_code);
        return err_code;
      }
    } else {
      // If not, we perform the operation right away.
      NRF_LOG_DEBUG("Performing Operation\n");
      memcpy(((void *)(&m_current_operation)), ((void *)(p_operation)), sizeof(m_current_operation));
      NRF_LOG_DEBUG("Operation Data\n");
      NRF_LOG_HEXDUMP_DEBUG(&m_current_operation.data, sizeof(m_current_operation.data));
      
      err_code = _storage_perform_current_operation();
      
      if (err_code) {
        NRF_LOG_ERROR("_storage_perform_current_operation fail: %lu\n",
                      err_code);
        return err_code;
      }      
    }
  } else {
    NRF_LOG_DEBUG("Pushing Operation in Queue\n");
    NRF_LOG_DEBUG("Operation Data\n");
    NRF_LOG_HEXDUMP_DEBUG(&(p_operation->data), sizeof(p_operation->data));
    
    err_code = nrf_queue_push(&m_operation_queue, p_operation);

    if (err_code) {
      NRF_LOG_ERROR("queue_push fail: %lu\n", err_code);
      return err_code;
    }    
  }

  return NRF_SUCCESS;
}

static void _storage_fds_evt_handler(fds_evt_t const *const p_fds_evt) {
  ret_code_t err_code = NRF_ERROR_INVALID_STATE;
  fds_evt_id_t id = p_fds_evt->id;

  if (id == FDS_EVT_INIT) {
    NRF_LOG_DEBUG("FDS Init Event\n");
    if (p_fds_evt->result != FDS_SUCCESS) {
      NRF_LOG_ERROR("FDS_EVT_INIT fail: %d\n", p_fds_evt->result);
      return;
    }
  } else if (id == FDS_EVT_WRITE || id == FDS_EVT_UPDATE ||
             id == FDS_EVT_DEL_RECORD) {
    if (id == FDS_EVT_WRITE || id == FDS_EVT_UPDATE) {
      is_writing = false;
      NRF_LOG_DEBUG("FDS Write or Update Event\n");
    } else if (id == FDS_EVT_DEL_RECORD) {
      is_deleting = false;
      NRF_LOG_DEBUG("FDS Delete Record Event\n");
    }

    NRF_LOG_DEBUG("Ret Code: %lu\r\n", p_fds_evt->result);
    // If the queue is not empty, fetch and perform the next operation. 
    if (!nrf_queue_is_empty(&m_operation_queue)) {
      err_code = nrf_queue_pop(&m_operation_queue, &m_current_operation);
      NRF_LOG_HEXDUMP_DEBUG(&m_current_operation.data, sizeof(m_current_operation.data));
      if (err_code) {
        NRF_LOG_ERROR("nrf_queue_pop: %d\n", err_code);
        return;
      }
      err_code = _storage_perform_current_operation();
      if (err_code) {
        NRF_LOG_ERROR("_storage_perform_current_operation: %d\n", err_code);
        return;
      }
    }
  } else if (id == FDS_EVT_GC) {
    is_collecting = false;
    NRF_LOG_DEBUG("FDS GC event\n");
    // Since fs_gc is called from _storage_perform_writing_operation, when there
    // is no memory left, we have to perform the current operation again.
    err_code = _storage_perform_current_operation();
    if (err_code) {
      NRF_LOG_ERROR("_storage_perform_current_operation: %d\n", err_code);
      return;
    }
  } else if (id == FDS_EVT_DEL_FILE) {
    NRF_LOG_DEBUG("FDS Delete File event\n");
  } else {
    NRF_LOG_ERROR("FDS Unknown Event: %d\n", p_fds_evt->id)
  }
}

ret_code_t storage_init(void) {
  NRF_LOG_DEBUG("storage_init\n");

  ret_code_t err_code;

  err_code = fds_register(_storage_fds_evt_handler);

  if (err_code) {
    NRF_LOG_ERROR("fds_register fail: %d\n", err_code);
    return err_code;
  }

  return NRF_SUCCESS;
}

ret_code_t storage_write_calibrations(uint16_t *p_calibrations) {
  NRF_LOG_DEBUG("Writing calibrations\n");

  storage_operation_t operation = {0};
  operation.type = STORAGE_WRITE_CALIBRATIONS_OPERATION;
  memcpy(operation.data.calibration_values, p_calibrations,
         sizeof(operation.data.calibration_values));

  nrf_delay_ms(10);
  NRF_LOG_DEBUG("Calibration Data\n");
  NRF_LOG_HEXDUMP_DEBUG(&operation.data, sizeof(operation.data));
  nrf_delay_ms(10);

  ret_code_t err_code = _storage_push_operation(&operation);

  if (err_code) {
    NRF_LOG_ERROR("_storage_push_operation fail: %d\n", err_code);
    return err_code;
  }

  return NRF_SUCCESS;
}

ret_code_t storage_read_calibrations(uint16_t *p_calibrations) {
  NRF_LOG_DEBUG("Reading calibrations\n");
  uint16_t length_words = 0;

  if ((sizeof(m_current_operation.data.calibration_values) %
       sizeof(uint32_t)) == 0) {
    length_words =
        sizeof(m_current_operation.data.calibration_values) / sizeof(uint32_t);
  } else {
    length_words = (sizeof(m_current_operation.data.calibration_values) /
                    sizeof(uint32_t)) +
                   1;
  }

  ret_code_t err_code = _storage_fds_read(STORAGE_CALIBRATIONS_KEY,
                                          p_calibrations, length_words);

  if (err_code) {
    if (err_code != FDS_ERR_NOT_FOUND) {
      NRF_LOG_ERROR("_storage_fds_read fail: %d\n", err_code);
    }
    return err_code;
  }

  return NRF_SUCCESS;
}

ret_code_t storage_delete_calibrations(void) {
  NRF_LOG_DEBUG("Deleting calibrations\n");

  storage_operation_t operation = {0};
  operation.type = STORAGE_DELETE_CALIBRATIONS_OPERATION;

  ret_code_t err_code = _storage_push_operation(&operation);

  if (err_code) {
    NRF_LOG_ERROR("_storage_push_operation fail: %d\n", err_code);
    return err_code;
  }

  return NRF_SUCCESS;
}

ret_code_t storage_write_dose_times(uint32_t *p_dose_times) {
  NRF_LOG_DEBUG("Writing Dose Times\n");

  storage_operation_t operation = {0};
  operation.type = STORAGE_WRITE_DOSE_TIMES_OPERATION;
  memcpy(operation.data.dose_times, p_dose_times,
         sizeof(operation.data.dose_times));

  nrf_delay_ms(10);
  NRF_LOG_DEBUG("Dose Times Data\n");
  NRF_LOG_HEXDUMP_DEBUG(&operation.data, sizeof(operation.data));
  nrf_delay_ms(10);

  ret_code_t err_code = _storage_push_operation(&operation);

  if (err_code) {
    NRF_LOG_ERROR("_storage_push_operation fail: %d\n", err_code);
    return err_code;
  }

  return NRF_SUCCESS;
}

ret_code_t storage_read_dose_times(uint32_t *p_dose_times) {
  NRF_LOG_DEBUG("Reading dose times\n")
  uint16_t length_words = 0;

  if ((sizeof(m_current_operation.data.dose_times) % sizeof(uint32_t)) == 0) {
    length_words =
        sizeof(m_current_operation.data.dose_times) / sizeof(uint32_t);
  } else {
    length_words =
        (sizeof(m_current_operation.data.dose_times) / sizeof(uint32_t)) + 1;
  }

  ret_code_t err_code =
      _storage_fds_read(STORAGE_DOSE_TIMES_KEY, p_dose_times, length_words);

  if (err_code) {
    if (err_code != FDS_ERR_NOT_FOUND) {
      NRF_LOG_ERROR("_storage_fds_read fail: %d\n", err_code);
    }
    return err_code;
  }

  return NRF_SUCCESS;
}

ret_code_t storage_delete_dose_times(void) {
  NRF_LOG_DEBUG("Deleting dose times\n");

  storage_operation_t operation = {0};
  operation.type = STORAGE_DELETE_DOSE_TIMES_OPERATION;

  ret_code_t err_code = _storage_push_operation(&operation);

  if (err_code) {
    NRF_LOG_ERROR("_storage_push_operation fail: %d\n", err_code);
    return err_code;
  }

  return NRF_SUCCESS;
}

ret_code_t storage_write_coulomb_counter(float charge, uint8_t battery_level) {
  NRF_LOG_DEBUG("Writing coulomb counter\n");

  storage_operation_t operation = {0};
  operation.type = STORAGE_WRITE_COULOMB_COUNTER_OPERATION;
  operation.data.coulomb_counter_data.charge = charge;
  operation.data.coulomb_counter_data.battery_level = battery_level;

  nrf_delay_ms(10);
  NRF_LOG_DEBUG("Coulomb Counter Data\n");
  NRF_LOG_HEXDUMP_DEBUG(&operation.data, sizeof(operation.data));
  nrf_delay_ms(10);


  ret_code_t err_code = _storage_push_operation(&operation);

  if (err_code) {
    NRF_LOG_ERROR("_storage_push_operation fail: %d\n", err_code);
    return err_code;
  }

  return NRF_SUCCESS;
}

ret_code_t storage_read_coulomb_counter(float *p_charge,
                                        uint8_t *p_battery_level) {
  NRF_LOG_DEBUG("Reading coulomb counter\n");
  storage_coulomb_counter_data_t data = {0};
  uint16_t length_words = 0;

  if ((sizeof(data) % sizeof(uint32_t)) == 0) {
    length_words = sizeof(data) / sizeof(uint32_t);
  } else {
    length_words = (sizeof(data) / sizeof(uint32_t)) + 1;
  }

  ret_code_t err_code =
      _storage_fds_read(STORAGE_CC_KEY, (void *)(&data), length_words);

  if (err_code) {
    if (err_code != FDS_ERR_NOT_FOUND) {
      NRF_LOG_ERROR("_storage_fds_read fail: %d\n", err_code);
    }
    return err_code;
  }

  *p_charge = data.charge;
  *p_battery_level = data.battery_level;

  return NRF_SUCCESS;
}

void storage_test(void) {
  float tx_cc_charge = 400.0;
  uint8_t tx_cc_battery_level = 50;

  uint16_t tx_calibrations[] = {
      14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1,
  };

  uint32_t tx_dose_times[] = {
      98, 97, 96, 95, 94, 93, 92, 91, 90, 89, 88, 87, 86, 85,
  };

  float rx_cc_charge = 0;
  uint8_t rx_cc_battery_level = 0;

  uint16_t rx_calibrations[14] = {0};
  uint32_t rx_dose_times[14] = {0};

  APP_ERROR_CHECK(storage_init());

  APP_ERROR_CHECK(
      storage_write_coulomb_counter(tx_cc_charge, tx_cc_battery_level));
  nrf_delay_ms(20);
  APP_ERROR_CHECK(storage_write_calibrations(tx_calibrations));
  nrf_delay_ms(20);
  APP_ERROR_CHECK(storage_write_dose_times(tx_dose_times));
  nrf_delay_ms(20);

  APP_ERROR_CHECK(
      storage_read_coulomb_counter(&rx_cc_charge, &rx_cc_battery_level));
  nrf_delay_ms(20);
  APP_ERROR_CHECK(storage_read_calibrations(rx_calibrations));
  nrf_delay_ms(20);
  APP_ERROR_CHECK(storage_read_dose_times(rx_dose_times));
  nrf_delay_ms(20);

  NRF_LOG_INFO("rx_cc_charge =" NRF_LOG_FLOAT_MARKER "\r\n",
               NRF_LOG_FLOAT(rx_cc_charge));
  NRF_LOG_INFO("rx_cc_battery_level = %d\n", rx_cc_battery_level);

  for (uint32_t i = 0; i < 14; i++) {
    NRF_LOG_INFO("rx_calibrations[%d] = %d\n", i, rx_calibrations[i]);
    nrf_delay_ms(10);
  }

  nrf_delay_ms(10);

  for (uint32_t i = 0; i < 14; i++) {
    NRF_LOG_INFO("rx_dose_times[%d] = %d\n", i, rx_dose_times[i]);
    nrf_delay_ms(10);
  }
}
